/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import au.com.bytecode.opencsv.CSVReader;
import dk.eobjects.metamodel.data.DataSet;
import dk.eobjects.metamodel.query.SelectItem;
import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.schema.ColumnType;
import dk.eobjects.metamodel.schema.Schema;
import dk.eobjects.metamodel.schema.Table;
import dk.eobjects.metamodel.schema.TableType;
import dk.eobjects.metamodel.util.FileHelper;

/**
 * DataContextStrategy to use for CSV files
 */
public class CsvDataContextStrategy extends QueryPostprocessDataContextStrategy {

	private File _file;
	private char _separatorChar;
	private char _quoteChar;

	public CsvDataContextStrategy(File file) {
		this(file, CSVReader.DEFAULT_SEPARATOR,
				CSVReader.DEFAULT_QUOTE_CHARACTER);
	}

	public CsvDataContextStrategy(File file, char separatorChar) {
		this(file, separatorChar, CSVReader.DEFAULT_QUOTE_CHARACTER);
	}

	/**
	 * Constructs a new CsvDataContextStrategy
	 * 
	 * @param file
	 *            the CSV file
	 * @param separatorChar
	 *            the character to use for seperating values
	 * @param quoteChar
	 *            the character to use for enclosing/quoting values
	 */
	public CsvDataContextStrategy(File file, char separatorChar, char quoteChar) {
		if (file == null) {
			throw new IllegalArgumentException("File cannot be null");
		}
		if (!file.exists() || !file.canRead()) {
			throw new IllegalArgumentException("Cannot read from file");
		}
		_file = file;
		_separatorChar = separatorChar;
		_quoteChar = quoteChar;
	}

	@Override
	public DataSet materializeTable(Table table, Column[] columns, int maxRows) {
		if (columns == null || columns.length == 0) {
			columns = table.getColumns();
		}
		CSVReader reader = null;
		try {
			reader = new CSVReader(FileHelper.getReader(_file), _separatorChar,
					_quoteChar, 1);

			List<Object[]> resultData = new ArrayList<Object[]>();
			for (String[] csvValues = reader.readNext(); csvValues != null; csvValues = reader
					.readNext()) {
				if (maxRows == 0) {
					break;
				}
				maxRows--;
				Object[] dataValues = new Object[columns.length];
				for (int i = 0; i < columns.length; i++) {
					Column column = columns[i];
					int columnNumber = column.getColumnNumber();
					if (columnNumber < csvValues.length) {
						dataValues[i] = csvValues[columnNumber];
					} else {
						// Ticket #125: Missing values should be enterpreted as
						// null.
						dataValues[i] = null;
					}
				}
				resultData.add(dataValues);
			}
			SelectItem[] selectItems = new SelectItem[columns.length];
			for (int i = 0; i < columns.length; i++) {
				Column column = columns[i];
				selectItems[i] = new SelectItem(column);
			}
			DataSet result = new DataSet(selectItems, resultData);
			return result;
		} catch (Exception e) {
			throw new IllegalStateException("Exception reading from file: "
					+ _file, e);
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					// do nothing
				}
			}
		}
	}

	public String getDefaultSchemaName() {
		String filename = _file.getName();
		return filename;
	}

	public Schema getSchemaByName(String name) throws MetaModelException {
		CSVReader reader = null;
		try {
			String schemaName = getDefaultSchemaName();
			Schema schema = new Schema(schemaName);
			Table table = new Table(schemaName.substring(0,
					schemaName.length() - 4), TableType.TABLE, schema);
			schema.addTable(table);

			reader = new CSVReader(FileHelper.getReader(_file), _separatorChar,
					_quoteChar);
			String[] columnHeaders = reader.readNext();
			reader.close();
			for (int i = 0; i < columnHeaders.length; i++) {
				Column column = new Column(columnHeaders[i],
						ColumnType.VARCHAR, table, i, true);
				table.addColumn(column);
			}
			table.setSchema(schema);
			return schema;
		} catch (Exception e) {
			throw new IllegalStateException("Exception reading from file: "
					+ _file, e);
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					// do nothing
				}
			}
		}
	}

	public String[] getSchemaNames() throws MetaModelException {
		return new String[] { getDefaultSchemaName() };
	}
}
