/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel;

import java.io.File;
import java.sql.Connection;
import java.util.Map;
import java.util.TreeMap;

import javax.sql.DataSource;

import org.apache.commons.lang.ArrayUtils;

import dk.eobjects.metamodel.data.DataSet;
import dk.eobjects.metamodel.query.Query;
import dk.eobjects.metamodel.schema.Schema;
import dk.eobjects.metamodel.schema.TableType;

/**
 * A DataContext represents the central entry point for interactions with
 * datastores. The DataContext contains of the structure of data (in the form of
 * schemas) and interactions (in the form of queries) with data.
 */
public class DataContext {

	public static final TableType[] DEFAULT_TABLE_TYPES = new TableType[] {
			TableType.TABLE, TableType.VIEW };
	private IDataContextStrategy _strategy;
	private Map<String, Schema> _schemas = new TreeMap<String, Schema>(
			SchemaNameComparator.getInstance());
	private String[] _schemaNames;

	/**
	 * Creates a new DataContext based on a datasource object
	 * 
	 * @param dataSource
	 */
	public DataContext(DataSource dataSource) {
		this(dataSource, DEFAULT_TABLE_TYPES);
	}

	/**
	 * Creates a new DataContext based on a datasource object and a catalog name
	 * 
	 * @param dataSource
	 * @param catalogName
	 */
	public DataContext(DataSource dataSource, String catalogName) {
		this(dataSource, catalogName, DEFAULT_TABLE_TYPES);
	}

	/**
	 * Creates a new DataContext based on a datasource object and desired table
	 * types
	 * 
	 * @param dataSource
	 * @param tableTypes
	 */
	public DataContext(DataSource dataSource, TableType... tableTypes) {
		this(dataSource, null, tableTypes);
	}

	/**
	 * Creates a new DataContext based on a datasource object, a catalog name
	 * and desired table types
	 * 
	 * @param dataSource
	 * @param catalogName
	 * @param tableTypes
	 */
	public DataContext(DataSource dataSource, String catalogName,
			TableType... tableTypes) {
		this(new JdbcDataContextStrategy(dataSource, tableTypes, catalogName));
	}

	/**
	 * Creates a new DataContext based on a JDBC connection
	 * 
	 * @param connection
	 */
	public DataContext(Connection connection) {
		this(connection, null, DEFAULT_TABLE_TYPES);
	}

	/**
	 * Creates a new DataContext based on a JDBC connection and a specified
	 * catalog name
	 * 
	 * @param connection
	 * @param catalogName
	 */
	public DataContext(Connection connection, String catalogName) {
		this(connection, catalogName, DEFAULT_TABLE_TYPES);
	}

	/**
	 * Creates a new DataContext based on a JDBC connection, only including
	 * tables of the specied types. Including the TableType.OTHER type will
	 * include all.
	 * 
	 * @param connection
	 * @param tableTypes
	 */
	public DataContext(Connection connection, TableType... tableTypes) {
		this(connection, null, tableTypes);
	}

	/**
	 * Creates a new DataContext based on a JDBC connection and a specified
	 * catalog name, only including tables of the specied types. Including the
	 * TableType.OTHER type will include all.
	 * 
	 * @param connection
	 * @param catalogName
	 * @param tableTypes
	 */
	public DataContext(Connection connection, String catalogName,
			TableType... tableTypes) {
		_strategy = new JdbcDataContextStrategy(connection, tableTypes,
				catalogName);
	}

	/**
	 * Creates a new DataContext based on a file. The file's filename (and in
	 * particular it's suffix) will be used to determine how to read the file,
	 * currently CSV, TSV, ODB, XML and XLS files are supported. For more
	 * configuration options (for example of CSV delimitors etc.) use the
	 * DataContext(IDataContextStrategy) constructor with an appropriate
	 * strategy.
	 * 
	 * @param file
	 *            a file containing data, either a Comma-separated file, an
	 *            OpenOffice.org database file or an Excel Spreadsheet file
	 * 
	 * @throws IllegalArgumentException
	 *             if the file does not have a recognized suffix
	 */
	public DataContext(File file) {
		if (file != null && file.exists() && file.isFile()) {
			String filename = file.getName().toLowerCase();
			if (filename.endsWith(".csv")) {
				_strategy = new CsvDataContextStrategy(file);
			} else if (filename.endsWith(".tsv")) {
				_strategy = new CsvDataContextStrategy(file, '\t');
			} else if (filename.endsWith(".xls")) {
				_strategy = new ExcelDataContextStrategy(file);
			} else if (filename.endsWith(".odb")) {
				_strategy = new OpenOfficeDataContextStrategy(file);
			} else if (filename.endsWith(".xml")) {
				_strategy = new XmlDataContextStrategy(file);
			} else {
				throw new IllegalArgumentException(
						"Cannot determine which strategy to use for opening file: "
								+ file);
			}
		} else {
			throw new IllegalArgumentException("Cannot read from file: " + file);
		}
	}

	/**
	 * Creates a DataContext based on a specified IDataContextStrategy. This
	 * enables development of custom datastore formats, by developing your own
	 * strategy-objects.
	 * 
	 * @param strategy
	 */
	public DataContext(IDataContextStrategy strategy) {
		_strategy = strategy;
	}

	/**
	 * Enforces a refresh of the schemas.
	 */
	public DataContext refreshSchemas() {
		_schemas.clear();
		_schemaNames = null;
		return this;
	}

	/**
	 * @return the schemas in this data context. Schemas are cached for reuse in
	 *         many situations so if you want to update the schemas, use the
	 *         refreshSchemas() method.
	 * @throws MetaModelException
	 *             if an error occurs retrieving the schema model
	 */
	public Schema[] getSchemas() throws MetaModelException {
		String[] schemaNames = getSchemaNames();
		for (int i = 0; i < schemaNames.length; i++) {
			String name = schemaNames[i];
			if (!_schemas.containsKey(name)) {
				Schema schema = getSchemaByName(name);
				_schemas.put(name, schema);
			}
		}
		Schema[] result = _schemas.values().toArray(new Schema[0]);
		return result;
	}

	/**
	 * @return an array of valid schema names
	 * @throws MetaModelException
	 *             if an error occurs retrieving the schema model
	 */
	public String[] getSchemaNames() throws MetaModelException {
		if (_schemaNames == null) {
			_schemaNames = _strategy.getSchemaNames();
		}
		return _schemaNames;
	}

	/**
	 * @return the schema that you are most probable to be interested in. The
	 *         default schema is determined by finding the schema with most
	 *         available of tables. In a lot of situations there will only be a
	 *         single available schema and in that case this will of course be
	 *         the schema returned.
	 * @throws MetaModelException
	 *             if an error occurs retrieving the schema model
	 */
	public Schema getDefaultSchema() throws MetaModelException {
		Schema result = null;
		String defaultSchemaName = _strategy.getDefaultSchemaName();
		if (defaultSchemaName != null) {
			result = getSchemaByName(defaultSchemaName);
		}
		if (result == null) {
			Schema[] schemas = getSchemas();
			if (schemas.length == 1) {
				result = schemas[0];
			} else {
				int highestTableCount = -1;
				for (int i = 0; i < schemas.length; i++) {
					Schema schema = schemas[i];
					if (schema.getTableCount() > highestTableCount) {
						highestTableCount = schema.getTableCount();
						result = schema;
					}
				}
			}
		}
		return result;
	}

	/**
	 * Gets a schema by a specified name
	 * 
	 * @param name
	 *            the name of the desired schema
	 * @return the Schema with the specified name or null if no such schema
	 *         exists
	 * @throws MetaModelException
	 *             if an error occurs retrieving the schema model
	 */
	public Schema getSchemaByName(String name) throws MetaModelException {
		Schema schema = _schemas.get(name);
		if (schema != null) {
			return schema;
		} else {
			if (ArrayUtils.indexOf(getSchemaNames(), name) != -1) {
				schema = _strategy.getSchemaByName(name);
				_schemas.put(name, schema);
			} else {
				// schema name does not exist
				schema = null;
			}
		}
		return schema;
	}

	/**
	 * Executes a query against the data context.
	 * 
	 * @param query
	 *            the query object to execute
	 * @return the dataset produced from executing the query
	 * @throws MetaModelException
	 *             if the specified query does not make sense or cannot be
	 *             executed because of restraints on the type of datastore.
	 */
	public DataSet executeQuery(Query query) throws MetaModelException {
		DataSet dataSet = _strategy.executeQuery(query);
		return dataSet;
	}
}