package dk.eobjects.metamodel;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;

import dk.eobjects.metamodel.data.DataSet;
import dk.eobjects.metamodel.query.SelectItem;
import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.schema.ColumnType;
import dk.eobjects.metamodel.schema.Schema;
import dk.eobjects.metamodel.schema.Table;
import dk.eobjects.metamodel.util.FormatHelper;

/**
 * DataContextStrategy to use for Excel files
 */
public class ExcelDataContextStrategy extends
		QueryPostprocessDataContextStrategy {

	private static final NumberFormat _numberFormat = FormatHelper
			.getUiNumberFormat();
	private File _file;

	public ExcelDataContextStrategy(File file) {
		if (file == null) {
			throw new IllegalArgumentException("File cannot be null");
		}
		if (!file.exists() || !file.canRead()) {
			throw new IllegalArgumentException("Cannot read from file");
		}
		_file = file;
	}

	@Override
	public DataSet materializeTable(Table table, Column[] columns, int maxRows) {
		if (columns == null || columns.length == 0)
			columns = table.getColumns();
		InputStream in = null;

		try {
			in = new FileInputStream(_file);
			HSSFWorkbook wb = new HSSFWorkbook(new POIFSFileSystem(in));
			HSSFSheet sheet = wb.getSheet(table.getName());
			// the first row is the Column names
			List<Object[]> resultData = new ArrayList<Object[]>();
			if (sheet != null && sheet.getPhysicalNumberOfRows() > 1) {
				for (int i = 1; i < sheet.getPhysicalNumberOfRows(); i++) {
					if (maxRows == 0) {
						break;
					}
					maxRows--;
					String[] dataValues = new String[columns.length];
					HSSFRow row = sheet.getRow(i);
					for (int j = 0; j < columns.length; j++) {
						HSSFCell cell = row.getCell((short) columns[j]
								.getColumnNumber());
						if (cell != null) {
							if (cell.getCellType() == HSSFCell.CELL_TYPE_NUMERIC) {
								cell.getNumericCellValue();
								dataValues[j] = _numberFormat.format(cell
										.getNumericCellValue());
							} else {
								dataValues[j] = cell.toString();
							}
						} else {
							dataValues[j] = "";
						}
					}
					resultData.add(dataValues);
				}
				SelectItem[] selectItems = new SelectItem[columns.length];
				for (int i = 0; i < columns.length; i++) {
					Column column = columns[i];
					selectItems[i] = new SelectItem(column);
				}
				DataSet result = new DataSet(selectItems, resultData);
				return result;

			}

		} catch (Exception e) {
			throw new IllegalStateException("Exception reading from file: "
					+ _file, e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
		}
		return null;
	}

	public String getDefaultSchemaName() {
		String fileName = _file.getName();
		return fileName;
	}

	public Schema getSchemaByName(String name) throws MetaModelException {
		InputStream in = null;
		Schema schema = null;
		try {

			schema = new Schema(getDefaultSchemaName());
			in = new FileInputStream(_file);
			HSSFWorkbook wb = new HSSFWorkbook(new POIFSFileSystem(in));

			for (int i = 0; i < wb.getNumberOfSheets(); i++) {
				HSSFSheet currentSheet = wb.getSheetAt(i);
				if (currentSheet.getPhysicalNumberOfRows() > 0) {
					Table table = new Table(wb.getSheetName(i));
					HSSFRow row = currentSheet.getRow(0);
					int physicalNumberOfCells = row.getPhysicalNumberOfCells();
					for (short j = 0; j < physicalNumberOfCells; j++) {
						HSSFCell cell = row.getCell(j);
						Column column = new Column(cell.toString(),
								ColumnType.VARCHAR, table, j, true);
						table.addColumn(column);
					}
					table.setSchema(schema);
					schema.addTable(table);
				}
			}

		} catch (Exception e) {
			throw new IllegalStateException("Exception reading from file: "
					+ _file, e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {

				}
			}
		}
		return schema;
	}

	public String[] getSchemaNames() throws MetaModelException {
		return new String[] { getDefaultSchemaName() };
	}

}
