/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import dk.eobjects.metamodel.data.DataSet;
import dk.eobjects.metamodel.data.Row;
import dk.eobjects.metamodel.query.FilterItem;
import dk.eobjects.metamodel.query.FromItem;
import dk.eobjects.metamodel.query.GroupByItem;
import dk.eobjects.metamodel.query.OrderByItem;
import dk.eobjects.metamodel.query.Query;
import dk.eobjects.metamodel.query.SelectItem;
import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.schema.Table;
import dk.eobjects.metamodel.util.ObjectComparator;

/**
 * This class contains various helper functionality to common tasks involving
 * MetaModel, eg.:
 * <ul>
 * <li>Easy-access for traversing common schema items</li>
 * <li>Manipulate data in memory. These methods are primarily used to enable
 * queries for non-queryable data sources like CSV files and spreadsheets.</li>
 * <li>Query rewriting, traversing and manipulation.</li>
 * </ul>
 */
public class MetaModelHelper {

	private static final Log _log = LogFactory.getLog(MetaModelHelper.class);

	// Preven instantiation
	private MetaModelHelper() {
	}

	/**
	 * Creates an array of tables where all occurences of tables in the provided
	 * list of tables and columns are included
	 */
	public static Table[] getTables(List<Table> tableList,
			List<Column> columnList) {
		HashSet<Table> set = new HashSet<Table>();
		set.addAll(tableList);
		for (Column column : columnList) {
			set.add(column.getTable());
		}
		return set.toArray(new Table[set.size()]);
	}

	/**
	 * Converts a list of columns to a corresponding array of tables
	 * 
	 * @param columns
	 *            the columns that the tables will be extracted from
	 * @return
	 */
	public static Table[] getTables(List<Column> columns) {
		ArrayList<Table> result = new ArrayList<Table>();
		for (Column column : columns) {
			Table table = column.getTable();
			if (!result.contains(table)) {
				result.add(table);
			}
		}
		return result.toArray(new Table[result.size()]);
	}

	/**
	 * Creates a subset array of columns, where only columns that are contained
	 * within the specified table are included.
	 * 
	 * @param table
	 * @param columns
	 * @return
	 */
	public static Column[] getTableColumns(Table table, List<Column> columns) {
		return getTableColumns(table, columns
				.toArray(new Column[columns.size()]));
	}

	/**
	 * Creates a subset array of columns, where only columns that are contained
	 * within the specified table are included.
	 * 
	 * @param table
	 * @param columns
	 * @return
	 */
	public static Column[] getTableColumns(Table table, Column[] columns) {
		ArrayList<Column> result = new ArrayList<Column>();
		for (int i = 0; i < columns.length; i++) {
			Column column = columns[i];
			if (table == column.getTable()) {
				result.add(column);
			}
		}
		return result.toArray(new Column[result.size()]);
	}

	public static DataSet getCarthesianProduct(DataSet... fromDataSets) {
		return getCarthesianProduct(fromDataSets, new FilterItem[0]);
	}

	public static DataSet getCarthesianProduct(DataSet[] fromDataSets,
			List<FilterItem> whereItems) {
		return getCarthesianProduct(fromDataSets, whereItems
				.toArray(new FilterItem[whereItems.size()]));
	}

	public static DataSet getCarthesianProduct(DataSet[] fromDataSets,
			FilterItem... filterItems) {
		List<SelectItem> selectItems = new ArrayList<SelectItem>();
		for (DataSet dataSet : fromDataSets) {
			for (int i = 0; i < dataSet.getSelectItems().length; i++) {
				SelectItem item = dataSet.getSelectItems()[i];
				selectItems.add(item);
			}
		}

		int selectItemOffset = 0;
		List<Object[]> data = new ArrayList<Object[]>();
		for (int fromDataSetIndex = 0; fromDataSetIndex < fromDataSets.length; fromDataSetIndex++) {
			DataSet fromDataSet = fromDataSets[fromDataSetIndex];
			SelectItem[] fromSelectItems = fromDataSet.getSelectItems();
			if (fromDataSetIndex == 0) {
				while (fromDataSet.next()) {
					Object[] values = fromDataSet.getRow().getValues();
					Object[] row = new Object[selectItems.size()];
					System.arraycopy(values, 0, row, selectItemOffset,
							values.length);
					data.add(row);
				}
				fromDataSet.close();
			} else {
				List<Object[]> fromDataRows = new ArrayList<Object[]>();
				while (fromDataSet.next()) {
					fromDataRows.add(fromDataSet.getRow().getValues());
				}
				fromDataSet.close();
				for (int i = 0; i < data.size(); i = i + fromDataRows.size()) {
					Object[] originalRow = data.get(i);
					data.remove(i);
					for (int j = 0; j < fromDataRows.size(); j++) {
						Object[] newRow = fromDataRows.get(j);
						System.arraycopy(newRow, 0, originalRow,
								selectItemOffset, newRow.length);
						data.add(i + j, originalRow.clone());
					}
				}
			}
			selectItemOffset += fromSelectItems.length;
		}

		SelectItem[] finalSelectItems = selectItems
				.toArray(new SelectItem[selectItems.size()]);
		for (Iterator<Object[]> it = data.iterator(); it.hasNext();) {
			Object[] objects = it.next();
			Row row = new Row(finalSelectItems, objects);
			for (FilterItem whereItem : filterItems) {
				if (!whereItem.evaluate(row)) {
					it.remove();
					break;
				}
			}
		}
		return new DataSet(finalSelectItems, data);
	}

	public static DataSet getFiltered(DataSet dataSet,
			List<FilterItem> filterItems) {
		return getFiltered(dataSet, filterItems
				.toArray(new FilterItem[filterItems.size()]));
	}

	public static DataSet getFiltered(DataSet dataSet,
			FilterItem... filterItems) {
		List<Object[]> data = new ArrayList<Object[]>();
		while (dataSet.next()) {
			Row row = dataSet.getRow();
			boolean valid = true;
			for (FilterItem havingItem : filterItems) {
				if (!havingItem.evaluate(row)) {
					valid = false;
					break;
				}
			}
			if (valid) {
				data.add(row.getValues());
			}
		}
		dataSet.close();
		return new DataSet(dataSet.getSelectItems(), data);
	}

	public static DataSet getSelection(List<SelectItem> selectItems,
			DataSet dataSet) {
		return getSelection(selectItems.toArray(new SelectItem[selectItems
				.size()]), dataSet);
	}

	public static DataSet getSelection(SelectItem[] selectItems, DataSet dataSet) {
		SelectItem[] dataSetSelectItems = dataSet.getSelectItems();
		if (!ArrayUtils.isEquals(dataSetSelectItems, selectItems)) {
			List<Object[]> resultData = new ArrayList<Object[]>();
			while (dataSet.next()) {
				Row row = dataSet.getRow();
				Object[] resultRow = new Object[selectItems.length];
				for (int i = 0; i < selectItems.length; i++) {
					SelectItem selectItem = selectItems[i];
					if (selectItem.getSubQuerySelectItem() != null) {
						resultRow[i] = row.getValue(selectItem
								.getSubQuerySelectItem());
					} else {
						resultRow[i] = row.getValue(selectItem);
					}
				}
				resultData.add(resultRow);
			}
			dataSet.close();
			dataSet = new DataSet(selectItems, resultData);
		}
		return dataSet;
	}

	public static DataSet getGrouped(List<SelectItem> selectItems,
			DataSet dataSet, List<GroupByItem> groupByItems) {
		return getGrouped(selectItems
				.toArray(new SelectItem[selectItems.size()]), dataSet,
				groupByItems.toArray(new GroupByItem[groupByItems.size()]));
	}

	public static DataSet getGrouped(SelectItem[] selectItems, DataSet dataSet,
			GroupByItem[] groupByItems) {
		DataSet result = dataSet;
		if (groupByItems != null && groupByItems.length > 0) {
			Map<Row, Map<SelectItem, List<Object>>> uniqueRows = new HashMap<Row, Map<SelectItem, List<Object>>>();

			SelectItem[] groupBySelects = new SelectItem[groupByItems.length];
			for (int i = 0; i < groupBySelects.length; i++) {
				groupBySelects[i] = groupByItems[i].getSelectItem();
			}

			// Creates a list of SelectItems that have functions
			List<SelectItem> functionItems = getFunctionSelectItems(Arrays
					.asList(selectItems));

			// Loop through the dataset and identify groups
			while (dataSet.next()) {
				Row row = dataSet.getRow();

				// Subselect a row prototype with only the unique values that
				// define the group
				Row uniqueRow = row.getSubSelection(groupBySelects);

				// function input is the values used for calculating aggregate
				// functions in the group
				Map<SelectItem, List<Object>> functionInput;
				if (!uniqueRows.containsKey(uniqueRow)) {
					// If this group already exist, use an existing function
					// input
					functionInput = new HashMap<SelectItem, List<Object>>();
					for (SelectItem item : functionItems) {
						functionInput.put(item, new ArrayList<Object>());
					}
					uniqueRows.put(uniqueRow, functionInput);
				} else {
					// If this is a new group, create a new function input
					functionInput = uniqueRows.get(uniqueRow);
				}

				// Loop through aggregate functions to check for validity
				for (SelectItem item : functionItems) {
					List<Object> objects = functionInput.get(item);
					Column column = item.getColumn();
					if (column != null) {
						Object value = row.getValue(new SelectItem(column));
						objects.add(value);
					} else if (SelectItem.isCountAllItem(item)) {
						// Just use the empty string, since COUNT(*) don't
						// evaluate values (but null values should be prevented)
						objects.add("");
					} else {
						throw new IllegalArgumentException(
								"Expression function not supported: " + item);
					}
				}
			}

			dataSet.close();
			List<Object[]> resultData = new ArrayList<Object[]>();

			// Loop through the groups to generate aggregates
			for (Entry<Row, Map<SelectItem, List<Object>>> entry : uniqueRows
					.entrySet()) {
				Row row = entry.getKey();
				Map<SelectItem, List<Object>> functionInput = entry.getValue();
				Object[] resultRow = new Object[selectItems.length];
				// Loop through select items to generate a row
				for (int i = 0; i < selectItems.length; i++) {
					SelectItem item = selectItems[i];
					int uniqueRowIndex = row.indexOf(item);
					if (uniqueRowIndex != -1) {
						// If there's already a value for the select item in the
						// row, keep it (it's one of the grouped by columns)
						resultRow[i] = row.getValue(uniqueRowIndex);
					} else {
						// Use the function input to calculate the aggregate
						// value
						List<Object> objects = functionInput.get(item);
						if (objects != null) {
							Object functionResult = item.getFunction()
									.evaluate(objects.toArray());
							resultRow[i] = functionResult;
						} else {
							if (item.getFunction() != null) {
								_log
										.error("No function input found for SelectItem: "
												+ item);
							}
						}
					}
				}
				resultData.add(resultRow);
			}
			result = new DataSet(selectItems, resultData);
		}
		result = getSelection(selectItems, result);
		return result;
	}

	public static DataSet getAggregated(List<SelectItem> resultSelectItems,
			DataSet dataSet) {
		List<SelectItem> functionItems = getFunctionSelectItems(resultSelectItems);
		if (functionItems.isEmpty()) {
			return dataSet;
		} else {
			SelectItem[] resultSelectItemsArray = resultSelectItems
					.toArray(new SelectItem[resultSelectItems.size()]);
			Map<SelectItem, List<Object>> functionInput = new HashMap<SelectItem, List<Object>>();
			for (SelectItem item : functionItems) {
				functionInput.put(item, new ArrayList<Object>());
			}

			boolean onlyAggregates = true;
			if (functionItems.size() != resultSelectItems.size()) {
				onlyAggregates = false;
			}

			List<Row> resultRows = new ArrayList<Row>();
			while (dataSet.next()) {
				Row inputRow = dataSet.getRow();
				for (SelectItem item : functionItems) {
					List<Object> objects = functionInput.get(item);
					Column column = item.getColumn();
					if (column != null) {
						Object value = inputRow
								.getValue(new SelectItem(column));
						objects.add(value);
					} else if (SelectItem.isCountAllItem(item)) {
						// Just use the empty string, since COUNT(*) don't
						// evaluate values (but null values should be prevented)
						objects.add("");
					} else {
						throw new IllegalArgumentException(
								"Expression function not supported: " + item);
					}
				}

				// If the result should also contain non-aggregated values, we
				// will keep those in the rows list
				if (!onlyAggregates) {
					Object[] values = new Object[resultSelectItemsArray.length];
					for (int i = 0; i < resultSelectItemsArray.length; i++) {
						Object value = inputRow
								.getValue(resultSelectItemsArray[i]);
						if (value != null) {
							values[i] = value;
						}
					}
					resultRows.add(new Row(resultSelectItemsArray, values));
				}
			}
			dataSet.close();

			// Create the aggregates
			Map<SelectItem, Object> functionResult = new HashMap<SelectItem, Object>();
			for (SelectItem item : functionItems) {
				List<Object> input = functionInput.get(item);
				Object result = item.getFunction().evaluate(input.toArray());
				functionResult.put(item, result);
			}
			functionInput = null;

			if (onlyAggregates) {
				// We will only create a single row with all the aggregates
				Object[] values = new Object[resultSelectItemsArray.length];
				for (int i = 0; i < resultSelectItemsArray.length; i++) {
					values[i] = functionResult.get(resultSelectItemsArray[i]);
				}
				Row row = new Row(resultSelectItemsArray, values);
				resultRows.add(row);
			} else {
				// We will create the aggregates as well as regular values
				for (int i = 0; i < resultRows.size(); i++) {
					Row row = resultRows.get(i);
					Object[] values = row.getValues();
					for (Entry<SelectItem, Object> entry : functionResult
							.entrySet()) {
						SelectItem item = entry.getKey();
						int itemIndex = row.indexOf(item);
						if (itemIndex != -1) {
							Object value = entry.getValue();
							values[itemIndex] = value;
						}
					}
					resultRows.set(i, new Row(resultSelectItemsArray, values));
				}
			}
			return new DataSet(resultRows);
		}
	}

	public static List<SelectItem> getFunctionSelectItems(
			List<SelectItem> selectItems) {
		List<SelectItem> result = new ArrayList<SelectItem>();
		for (SelectItem selectItem : selectItems) {
			if (selectItem.getFunction() != null) {
				result.add(selectItem);
			}
		}
		if (_log.isDebugEnabled()) {
			_log.debug("SelectItems with functions: " + result);
		}
		return result;
	}

	public static DataSet getOrdered(DataSet dataSet,
			List<OrderByItem> orderByItems) {
		return getOrdered(dataSet, orderByItems
				.toArray(new OrderByItem[orderByItems.size()]));
	}

	public static DataSet getOrdered(DataSet dataSet,
			final OrderByItem... orderByItems) {
		if (orderByItems != null && orderByItems.length != 0) {
			final int[] sortIndexes = new int[orderByItems.length];
			for (int i = 0; i < orderByItems.length; i++) {
				OrderByItem item = orderByItems[i];
				int indexOf = dataSet.indexOf(item.getSelectItem());
				sortIndexes[i] = indexOf;
			}
			List<Object[]> data = dataSet.toObjectArrays();

			final Comparator<Object> valueComparator = ObjectComparator
					.getComparator();

			Collections.sort(data, new Comparator<Object[]>() {
				public int compare(Object[] o1, Object[] o2) {
					for (int i = 0; i < sortIndexes.length; i++) {
						Object sortObj1 = o1[sortIndexes[i]];
						Object sortObj2 = o2[sortIndexes[i]];
						int compare = valueComparator.compare(sortObj1,
								sortObj2);
						if (compare != 0) {
							OrderByItem orderByItem = orderByItems[i];
							boolean ascending = orderByItem.isAscending();
							if (ascending) {
								return compare;
							} else {
								return compare * -1;
							}
						}
					}
					return 0;
				}
			});

			dataSet = new DataSet(dataSet.getSelectItems(), data);
		}
		return dataSet;
	}

	/**
	 * Examines a query and extracts an array of FromItem's that refer
	 * (directly) to tables (hence Joined FromItems and SubQuery FromItems are
	 * traversed but not included).
	 * 
	 * @param q
	 *            the query to examine
	 * @return
	 */
	public static FromItem[] getTableFromItems(Query q) {
		List<FromItem> result = new ArrayList<FromItem>();
		List<FromItem> items = q.getFromClause().getItems();
		for (FromItem item : items) {
			result.addAll(getTableFromItems(item));
		}
		return result.toArray(new FromItem[result.size()]);
	}

	public static List<FromItem> getTableFromItems(FromItem item) {
		List<FromItem> result = new ArrayList<FromItem>();
		if (item.getTable() != null) {
			result.add(item);
		} else if (item.getSubQuery() != null) {
			FromItem[] sqItems = getTableFromItems(item.getSubQuery());
			for (int i = 0; i < sqItems.length; i++) {
				result.add(sqItems[i]);
			}
		} else if (item.getJoin() != null) {
			FromItem leftSide = item.getLeftSide();
			result.addAll(getTableFromItems(leftSide));
			FromItem rightSide = item.getRightSide();
			result.addAll(getTableFromItems(rightSide));
		} else {
			throw new IllegalStateException(
					"FromItem was neither of Table type, SubQuery type or Join type: "
							+ item);
		}
		return result;
	}

	/**
	 * Executes a single row query, like "SELECT COUNT(*), MAX(SOME_COLUMN) FROM
	 * MY_TABLE" or similar.
	 * 
	 * @param dataContext
	 * @param query
	 * @return
	 * @throws MetaModelException
	 *             if less or more than one Row is returned from the query
	 */
	public static Row executeSingleRowQuery(DataContext dataContext, Query query)
			throws MetaModelException {
		DataSet dataSet = dataContext.executeQuery(query);
		boolean next = dataSet.next();
		if (!next) {
			throw new MetaModelException("No rows returned from query: "
					+ query);
		}
		Row row = dataSet.getRow();
		next = dataSet.next();
		if (next) {
			throw new MetaModelException(
					"More than one row returned from query: " + query);
		}
		dataSet.close();
		return row;
	}

	/**
	 * Performs a left join (aka left outer join) operation on two datasets.
	 * 
	 * @param ds1
	 *            the left dataset
	 * @param ds2
	 *            the right dataset
	 * @param onConditions
	 *            the conditions to join by
	 * @return the left joined result dataset
	 */
	public static DataSet getLeftJoin(DataSet ds1, DataSet ds2,
			FilterItem[] onConditions) {
		SelectItem[] si1 = ds1.getSelectItems();
		SelectItem[] si2 = ds2.getSelectItems();
		SelectItem[] selectItems = new SelectItem[si1.length + si2.length];
		System.arraycopy(si1, 0, selectItems, 0, si1.length);
		System.arraycopy(si2, 0, selectItems, si1.length, si2.length);

		List<Object[]> resultRows = new ArrayList<Object[]>();
		List<Object[]> ds2data = ds2.toObjectArrays();
		while (ds1.next()) {

			// Construct a single-row dataset for making a carthesian product
			// against ds2
			Row ds1row = ds1.getRow();
			List<Row> ds1rows = new ArrayList<Row>();
			ds1rows.add(ds1row);

			List<Object[]> carthesianRows = getCarthesianProduct(
					new DataSet[] { new DataSet(ds1rows),
							new DataSet(si2, ds2data) }, onConditions)
					.toObjectArrays();
			if (carthesianRows.size() > 0) {
				resultRows.addAll(carthesianRows);
			} else {
				Object[] values = ds1row.getValues();
				Object[] row = new Object[selectItems.length];
				System.arraycopy(values, 0, row, 0, values.length);
				resultRows.add(row);
			}
		}
		ds1.close();

		return new DataSet(selectItems, resultRows);
	}

	/**
	 * Performs a right join (aka right outer join) operation on two datasets.
	 * 
	 * @param ds1
	 *            the left dataset
	 * @param ds2
	 *            the right dataset
	 * @param onConditions
	 *            the conditions to join by
	 * @return the right joined result dataset
	 */
	public static DataSet getRightJoin(DataSet ds1, DataSet ds2,
			FilterItem[] onConditions) {
		SelectItem[] ds1selects = ds1.getSelectItems();
		SelectItem[] ds2selects = ds2.getSelectItems();
		SelectItem[] leftOrderedSelects = new SelectItem[ds1selects.length
				+ ds2selects.length];
		System.arraycopy(ds1selects, 0, leftOrderedSelects, 0,
				ds1selects.length);
		System.arraycopy(ds2selects, 0, leftOrderedSelects, ds1selects.length,
				ds2selects.length);

		// We will reuse the left join algorithm (but switch the datasets
		// around)
		DataSet dataSet = getLeftJoin(ds2, ds1, onConditions);
		
		dataSet = getSelection(leftOrderedSelects, dataSet);
		return dataSet;
	}
}
