/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import dk.eobjects.metamodel.data.DataSet;
import dk.eobjects.metamodel.query.Query;
import dk.eobjects.metamodel.schema.Schema;
import dk.eobjects.metamodel.util.FileHelper;

public class OpenOfficeDataContextStrategy implements IDataContextStrategy {

	private static final Log _log = LogFactory
			.getLog(OpenOfficeDataContextStrategy.class);
	private IDataContextStrategy _strategy;
	private Connection _connection;

	public OpenOfficeDataContextStrategy(File dbFile) throws MetaModelException {
		try {
			String databaseName = dbFile.getName();
			File tempDir = FileHelper.getTempDir();

			ZipFile zipFile = new ZipFile(dbFile);
			Enumeration<? extends ZipEntry> entries = zipFile.entries();
			while (entries.hasMoreElements()) {
				ZipEntry entry = entries.nextElement();
				String entryName = entry.getName();
				if (entryName.startsWith("database/")) {
					File destFile = new File(tempDir, databaseName + "."
							+ entryName.substring(9));
					destFile.createNewFile();

					if (_log.isDebugEnabled()) {
						_log.debug("Processing entry: " + entryName);
						_log.debug("Writing temp file: "
								+ destFile.getAbsolutePath());
					}

					byte[] buffer = new byte[1024];
					InputStream inputStream = zipFile.getInputStream(entry);
					OutputStream outputStream = new BufferedOutputStream(
							new FileOutputStream(destFile));
					for (int len = inputStream.read(buffer); len >= 0; len = inputStream
							.read(buffer)) {
						outputStream.write(buffer, 0, len);
					}
					inputStream.close();
					outputStream.close();
					destFile.deleteOnExit();
				} else {
					_log.debug("Ignoring entry: " + entryName);
				}
			}
			zipFile.close();

			Class.forName("org.hsqldb.jdbcDriver");
			String url = "jdbc:hsqldb:file:" + tempDir.getAbsolutePath() + "/"
					+ databaseName;
			_log.info("Using database URL: " + url);
			_connection = DriverManager.getConnection(url, "SA", "");
			_connection.setReadOnly(true);
			_strategy = new JdbcDataContextStrategy(_connection,
					DataContext.DEFAULT_TABLE_TYPES, null);
		} catch (Exception e) {
			throw new MetaModelException(e);
		}
	}

	public DataSet executeQuery(Query query) throws MetaModelException {
		return _strategy.executeQuery(query);
	}

	public String getDefaultSchemaName() throws MetaModelException {
		return _strategy.getDefaultSchemaName();
	}

	public Schema getSchemaByName(String name) throws MetaModelException {
		return _strategy.getSchemaByName(name);
	}

	public String[] getSchemaNames() throws MetaModelException {
		return _strategy.getSchemaNames();
	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		_connection.close();
	}

	public Connection getConnection() {
		return _connection;
	}
}
