/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.data;

import java.io.Closeable;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import javax.swing.table.TableModel;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import dk.eobjects.metamodel.MetaModelException;
import dk.eobjects.metamodel.query.Query;
import dk.eobjects.metamodel.query.SelectItem;

public class DataSet implements Closeable {

	private IDataSetStrategy _strategy;

	public DataSet(Query query, Statement statement, ResultSet resultSet) {
		this(new JdbcDataSetStrategy(query, statement, resultSet));
	}

	public DataSet(SelectItem[] selectItems, Statement statement,
			ResultSet resultSet) {
		this(new JdbcDataSetStrategy(selectItems, statement, resultSet));
	}

	public DataSet(IDataSetStrategy strategy) {
		if (strategy == null) {
			throw new IllegalArgumentException("Strategy cannot be null");
		}
		_strategy = strategy;
	}

	public DataSet(SelectItem[] selectItems, List<Object[]> data) {
		this(new InMemoryDataSetStrategy(selectItems, data, null));
	}

	public DataSet(List<Row> rows) {
		this(new InMemoryDataSetStrategy(rows));
	}

	public SelectItem[] getSelectItems() {
		return _strategy.getSelectItems();
	}

	public int indexOf(SelectItem item) {
		if (item != null) {
			SelectItem[] items = _strategy.getSelectItems();
			for (int i = 0; i < items.length; i++) {
				if (item.equalsIgnoreAlias(items[i])) {
					return i;
				}
			}
		}
		return -1;
	}

	public boolean next() {
		return _strategy.next();
	}

	public Row getRow() throws MetaModelException {
		return _strategy.getRow();
	}

	public void close() {
		_strategy.close();
	}

	/**
	 * Converts the DataSet into a TableModel
	 */
	public TableModel toTableModel() {
		TableModel tableModel = new DataSetTableModel(this);
		return tableModel;
	}

	/**
	 * Converts the DataSet into a list of object arrays
	 */
	public List<Object[]> toObjectArrays() {
		List<Object[]> objects = new ArrayList<Object[]>();
		while (next()) {
			Row row = getRow();
			objects.add(row.getValues());
		}
		close();
		return objects;
	}

	@Override
	public String toString() {
		return "DataSet[strategy=" + _strategy + "]";
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof DataSet) {
			DataSet that = (DataSet) obj;
			EqualsBuilder eb = new EqualsBuilder();
			eb.append(this._strategy, that._strategy);
			return eb.isEquals();
		}
		return false;
	}

	@Override
	public int hashCode() {
		return new HashCodeBuilder().append(_strategy).toHashCode();
	}
}