/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.dialects;

import java.util.List;

import dk.eobjects.metamodel.JdbcDataContextStrategy;
import dk.eobjects.metamodel.query.FromItem;
import dk.eobjects.metamodel.query.Query;
import dk.eobjects.metamodel.query.SelectItem;

/**
 * Generic query rewriter that adds syntax enhancements that are only possible
 * to resolve just before execution time.
 */
public class DefaultQueryRewriter extends AbstractQueryRewriter {

	private static final String SPECIAL_ALIAS_CHARACTERS = "- ,.|*%()!#¤/\\=?;:~";

	@Override
	protected Query beforeRewrite(JdbcDataContextStrategy strategy, Query query) {
		query = query.clone();

		if (strategy != null) {
			String identifierQuoteString = strategy.getIdentifierQuoteString();
			if (identifierQuoteString != null) {
				List<SelectItem> selectItems = query.getSelectClause()
						.getItems();
				for (SelectItem item : selectItems) {
					String alias = item.getAlias();
					if (needsQuoting(alias, identifierQuoteString)) {
						item.setAlias(identifierQuoteString + alias
								+ identifierQuoteString);
					}
				}
				List<FromItem> fromItems = query.getFromClause().getItems();
				for (FromItem item : fromItems) {
					String alias = item.getAlias();
					if (needsQuoting(alias, identifierQuoteString)) {
						item.setAlias(identifierQuoteString + alias
								+ identifierQuoteString);
					}
				}
			}
		}
		return query;
	}

	private boolean needsQuoting(String alias, String identifierQuoteString) {
		boolean result = false;
		if (alias != null && identifierQuoteString != null) {
			if (alias.indexOf(identifierQuoteString) == -1) {
				for (int i = 0; i < SPECIAL_ALIAS_CHARACTERS.length(); i++) {
					char specialCharacter = SPECIAL_ALIAS_CHARACTERS.charAt(i);
					if (alias.indexOf(specialCharacter) != -1) {
						result = true;
						break;
					}
				}
			}
		}
		if (_log.isDebugEnabled()) {
			_log.debug("needsQuoting(" + alias + "," + identifierQuoteString
					+ ") = " + result);
		}
		return result;
	}
}
