/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.query;

import java.text.NumberFormat;
import java.util.Comparator;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;

import dk.eobjects.metamodel.data.Row;
import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.schema.ColumnType;
import dk.eobjects.metamodel.util.FormatHelper;
import dk.eobjects.metamodel.util.ObjectComparator;

/**
 * Represents a filter in a query that resides either within a WHERE clause or a
 * HAVING clause
 * 
 * @see FilterClause
 * @see OperatorType
 */
public class FilterItem implements IQueryItem, Cloneable {

	private static final long serialVersionUID = 2435322742894653227L;
	private Query _query;
	private SelectItem _selectItem;
	private OperatorType _operator;
	private Object _operand;
	private FilterItem[] _items;

	/**
	 * Private constructor, used for cloning
	 */
	private FilterItem() {
	}

	/**
	 * Creates a single filter item based on a SelectItem, an operator and an
	 * operand.
	 * 
	 * @param selectItem
	 *            the selectItem to put constraints on, cannot be null
	 * @param operator
	 *            The operator to use. Can be OperatorType.EQUALS_TO,
	 *            OperatorType.DIFFERENT_FROM,
	 *            OperatorType.GREATER_THAN,OperatorType.SMALLER_THAN
	 * @param operand
	 *            The operand. Can be a constant like null or a String, a
	 *            Number, a Boolean, a Date, a Time, a DateTime. Or another
	 *            SelectItem
	 * @throws IllegalArgumentException
	 *             if the SelectItem is null or if the combination of operator
	 *             and operand does not make sense.
	 */
	public FilterItem(SelectItem selectItem, OperatorType operator,
			Object operand) throws IllegalArgumentException {
		_selectItem = selectItem;
		_operator = operator;
		_operand = operand;
		validate();
	}

	/**
	 * Creates a composite filter item based on other filter items. Each
	 * provided filter items will be OR'ed meaning that if one of the evaluates
	 * as true, then the composite filter will be evaluated as true
	 * 
	 * @param items
	 *            the items to include in the composite
	 */
	public FilterItem(FilterItem... items) {
		_items = items;
	}

	private void validate() {
		require("SelectItem cannot be null", _selectItem != null);
		if (_operand == null) {
			require(
					"Can only use EQUALS or DIFFERENT_FROM operator with null-operand",
					_operator == OperatorType.DIFFERENT_FROM
							|| _operator == OperatorType.EQUALS_TO);
		}
		if (_operator == OperatorType.LIKE) {
			ColumnType type = _selectItem.getColumn().getType();
			if (type != null) {
				require(
						"Can only use LIKE operator with strings",
						type.isLiteral()
								&& (_operand instanceof String || _operand instanceof SelectItem));
			}
		}
		if (_operand instanceof String) {
			String operandValue = (String) _operand;
			if (operandValue.indexOf('\'') != -1) {
				throw new UnsupportedOperationException(
						"String with single quote (') chars are not supported: "
								+ operandValue);
			}
		}
	}

	private void require(String errorMessage, boolean b) {
		if (!b) {
			throw new IllegalArgumentException(errorMessage);
		}
	}

	public SelectItem getSelectItem() {
		return _selectItem;
	}

	public FilterItem setSelectItem(SelectItem selectItem) {
		_selectItem = selectItem;
		validate();
		return this;
	}

	public OperatorType getOperator() {
		return _operator;
	}

	public FilterItem setOperator(OperatorType operator) {
		_operator = operator;
		validate();
		return this;
	}

	public Object getOperand() {
		return _operand;
	}

	public FilterItem setOperand(Object operand) {
		_operand = operand;
		validate();
		return this;
	}

	public Query getQuery() {
		return _query;
	}

	public FilterItem setQuery(Query query) {
		_query = query;
		if (_items == null) {
			if (_selectItem.getQuery() == null) {
				_selectItem.setQuery(_query);
			}
			if (_operand instanceof SelectItem) {
				SelectItem operand = (SelectItem) _operand;
				if (operand.getQuery() == null) {
					operand.setQuery(_query);
				}
			}
		} else {
			for (int i = 0; i < _items.length; i++) {
				FilterItem item = _items[i];
				if (item.getQuery() == null) {
					item.setQuery(_query);
				}
			}
		}
		return this;
	}

	/**
	 * Parses the constraint as a SQL Where-clause
	 */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		if (_items == null) {
			sb.append(_selectItem.getSameQueryAlias());
			if (_operand == null) {
				if (_operator == OperatorType.EQUALS_TO) {
					sb.append(" IS NULL");
				} else {
					sb.append(" IS NOT NULL");
				}
			} else {
				if (_operator == OperatorType.DIFFERENT_FROM) {
					sb.append(" <> ");
				} else if (_operator == OperatorType.EQUALS_TO) {
					sb.append(" = ");
				} else if (_operator == OperatorType.HIGHER_THAN) {
					sb.append(" > ");
				} else if (_operator == OperatorType.LIKE) {
					sb.append(" LIKE ");
				} else if (_operator == OperatorType.LOWER_THAN) {
					sb.append(" < ");
				} else {
					throw new IllegalStateException(
							"Operator could not be determined");
				}
				if (_operand instanceof SelectItem) {
					sb.append(((SelectItem) _operand).getSameQueryAlias());
				} else if (_operand instanceof String) {
					sb.append('\'' + (String) _operand + '\'');
				} else if (_operand instanceof Number) {
					NumberFormat numberFormat = FormatHelper
							.getSqlNumberFormat();
					String formattedOperand = numberFormat.format(_operand);
					sb.append(formattedOperand);
				} else if (_operand instanceof Boolean) {
					if ((Boolean) _operand) {
						sb.append("1");
					} else {
						sb.append("0");
					}
				} else if (_operand instanceof java.sql.Date
						|| _operand instanceof java.util.Date
						|| _operand instanceof java.sql.Time
						|| _operand instanceof java.util.Calendar
						|| _operand instanceof org.joda.time.LocalDate
						|| _operand instanceof org.joda.time.LocalTime
						|| _operand instanceof org.joda.time.DateTime) {
					ColumnType columnType = _selectItem.getColumn().getType();

					if (columnType == ColumnType.DATE) {
						LocalDate localDate = new LocalDate(_operand);
						sb.append("DATE '" + localDate.toString("yyyy-MM-dd")
								+ "'");
					} else if (columnType == ColumnType.TIME) {
						LocalTime localTime = new LocalTime(_operand);
						sb.append("TIME '"
								+ localTime.toString("HH:mm:ss.SSSSSS") + "'");
					} else if (columnType == ColumnType.TIMESTAMP) {
						DateTime dateTime = new DateTime(_operand);
						sb.append("TIMESTAMP '"
								+ dateTime
										.toString("yyyy-MM-dd HH:mm:ss.SSSSSS")
								+ "'");
					} else {
						throw new IllegalStateException(
								"Operand was date-type, but SelectItem's column was not.");
					}
				} else {
					throw new IllegalStateException(
							"Operand type not supported: "
									+ _operand.getClass());
				}
			}
		} else {
			sb.append('(');
			for (int i = 0; i < _items.length; i++) {
				FilterItem item = _items[i];
				if (i != 0) {
					sb.append(" OR ");
				}
				sb.append(item.toString());
			}
			sb.append(')');
		}

		return sb.toString();
	}

	/**
	 * Does a "manual" evaluation, useful for CSV data and alike, where queries
	 * cannot be created.
	 */
	@SuppressWarnings("unchecked")
	public boolean evaluate(Row row) {
		if (_items == null) {
			// Evaluate a single constraint
			Comparable selectItemValue = (Comparable) row.getValue(_selectItem);
			Object operandValue = _operand;
			if (_operand instanceof SelectItem) {
				operandValue = row.getValue((SelectItem) _operand);
			}
			if (operandValue == null) {
				if (_operator == OperatorType.DIFFERENT_FROM) {
					return (selectItemValue != null);
				} else if (_operator == OperatorType.EQUALS_TO) {
					return (selectItemValue == null);
				} else {
					return false;
				}
			} else if (selectItemValue == null) {
				if (_operator == OperatorType.DIFFERENT_FROM) {
					return true;
				} else {
					return false;
				}
			} else {
				Comparator comparator = ObjectComparator.getComparator();

				if (_operator == OperatorType.DIFFERENT_FROM) {
					return comparator.compare(selectItemValue, operandValue) != 0;
				} else if (_operator == OperatorType.EQUALS_TO) {
					return comparator.compare(selectItemValue, operandValue) == 0;
				} else if (_operator == OperatorType.HIGHER_THAN) {
					return comparator.compare(selectItemValue, operandValue) > 0;
				} else if (_operator == OperatorType.LOWER_THAN) {
					return comparator.compare(selectItemValue, operandValue) < 0;
				} else if (_operator == OperatorType.LIKE) {
					WildcardPattern matcher = new WildcardPattern(
							(String) operandValue, '%');
					return matcher.matches((String) selectItemValue);
				} else {
					throw new IllegalStateException(
							"Operator could not be determined");
				}
			}
		} else {
			// Evaluate several constraints
			for (int i = 0; i < _items.length; i++) {
				FilterItem item = _items[i];
				boolean result = item.evaluate(row);
				if (result) {
					return true;
				}
			}
			return false;
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof FilterItem) {
			FilterItem that = (FilterItem) obj;
			EqualsBuilder eb = new EqualsBuilder();
			eb.append(this._operator, that._operator);
			eb.append(this._operand, that._operand);
			eb.append(this._selectItem, that._selectItem);
			return eb.isEquals();
		}
		return false;
	}

	@Override
	public int hashCode() {
		HashCodeBuilder hashCodeBuilder = new HashCodeBuilder();
		hashCodeBuilder.append(_selectItem);
		hashCodeBuilder.append(_operator);
		hashCodeBuilder.append(_operand);
		return hashCodeBuilder.toHashCode();
	}

	@Override
	protected FilterItem clone() {
		FilterItem f = new FilterItem();
		if (_items != null) {
			f._items = _items.clone();
		}
		if (_operand != null) {
			if (_operand instanceof SelectItem) {
				f._operand = ((SelectItem) _operand).clone();
			} else {
				f._operand = _operand;
			}
		}
		if (_selectItem != null) {
			f._selectItem = _selectItem.clone();
		}
		f._operator = _operator;
		return f;
	}

	public boolean isReferenced(Column column) {
		if (column != null) {
			if (_selectItem != null) {
				if (_selectItem.isReferenced(column)) {
					return true;
				}
			}
			if (_operand != null && _operand instanceof SelectItem) {
				if (((SelectItem) _operand).isReferenced(column)) {
					return true;
				}
			}
			if (_items != null) {
				for (FilterItem item : _items) {
					if (item.isReferenced(column)) {
						return true;
					}
				}
			}
		}
		return false;
	}
}
