/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.query;

import java.util.Comparator;

import org.apache.commons.math.MathException;
import org.apache.commons.math.util.DefaultTransformer;

import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.util.ObjectComparator;

/**
 * Represents an aggregate function to use in a SelectItem.
 * 
 * @see SelectItem
 */
public enum FunctionType {

	COUNT, AVG, SUM, MAX, MIN;

	public SelectItem createSelectItem(Column column) {
		return new SelectItem(this, column);
	}

	public SelectItem createSelectItem(String expression, String alias) {
		return new SelectItem(this, expression, alias);
	}

	/**
	 * Executes the function
	 * 
	 * @param values
	 *            the values to be evaluated. If a value is null it won't be
	 *            evaluated
	 * @return
	 */
	public Object evaluate(Object... values) {
		if (this == COUNT) {
			long count = 0;
			for (Object value : values) {
				if (value != null) {
					count++;
				}
			}
			return count;
		}
		if (this == AVG) {
			double sum = 0.0;
			long count = 0;
			for (Object value : values) {
				if (value != null) {
					try {
						sum += new DefaultTransformer().transform(value);
						count++;
					} catch (MathException e) {
						throw new UnsupportedOperationException(
								"Cannot evaluate " + this
										+ " with argument type: "
										+ value.getClass());
					}
				}
			}
			return (sum / count);
		}
		if (this == SUM) {
			double sum = 0.0;
			for (Object value : values) {
				if (value != null) {
					try {
						sum += new DefaultTransformer().transform(value);
					} catch (MathException e) {
						throw new UnsupportedOperationException(
								"Cannot evaluate " + this
										+ " with argument type: "
										+ value.getClass());
					}
				}
			}
			return sum;
		}
		if (this == MAX) {
			Comparator<Object> comparator = ObjectComparator.getComparator();
			Object max = null;
			for (Object value : values) {
				if (value != null) {
					if (max == null) {
						max = value;
					}
					if (comparator.compare(value, max) > 0) {
						max = value;
					}
				}
			}
			return max;
		}
		if (this == MIN) {
			Comparator<Object> comparator = ObjectComparator.getComparator();
			Object min = null;
			for (Object value : values) {
				if (value != null) {
					if (min == null) {
						min = value;
					}
					if (comparator.compare(value, min) < 0) {
						min = value;
					}
				}
			}
			return min;
		}
		throw new UnsupportedOperationException("Cannot evaluate " + this);
	}
}
