/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.query;

import org.apache.commons.lang.builder.EqualsBuilder;

/**
 * Represents a GROUP BY item. GroupByItems always use a select item (that may
 * or not be a part of the query already) for grouping.
 * 
 * @see GroupByClause
 */
public class GroupByItem implements IQueryItem, Cloneable {

	private static final long serialVersionUID = 5218878395877852919L;
	private SelectItem _selectItem;
	private Query _query;

	/**
	 * Private constructor, used for cloning
	 */
	private GroupByItem() {
	}

	/**
	 * Constructs a GROUP BY item based on a select item that should be grouped.
	 * 
	 * @param selectItem
	 */
	public GroupByItem(SelectItem selectItem) {
		if (selectItem == null) {
			throw new IllegalArgumentException("SelectItem cannot be null");
		}
		_selectItem = selectItem;
	}

	public SelectItem getSelectItem() {
		return _selectItem;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(_selectItem.getSameQueryAlias());
		return sb.toString();
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == this) {
			return true;
		}
		if (obj instanceof GroupByItem) {
			GroupByItem that = (GroupByItem) obj;
			EqualsBuilder eb = new EqualsBuilder();
			eb.append(this.getSelectItem(), that.getSelectItem());
			return eb.isEquals();
		}
		return false;
	}

	public Query getQuery() {
		return _query;
	}

	public GroupByItem setQuery(Query query) {
		_query = query;
		if (_selectItem != null) {
			_selectItem.setQuery(query);
		}
		return this;
	}

	@Override
	protected GroupByItem clone() {
		GroupByItem g = new GroupByItem();
		g._selectItem = _selectItem.clone();
		return g;
	}
}
