/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.schema;

import java.lang.reflect.Field;
import java.util.Comparator;

import dk.eobjects.metamodel.util.NumberComparator;
import dk.eobjects.metamodel.util.ObjectComparator;
import dk.eobjects.metamodel.util.TimeComparator;
import dk.eobjects.metamodel.util.ToStringComparator;

/**
 * Represents the data-type of columns
 */
public enum ColumnType {

	/**
	 * Literal
	 */
	CHAR, VARCHAR, LONGVARCHAR, CLOB, NCHAR, NVARCHAR, LONGNVARCHAR, NCLOB,

	/**
	 * Numbers
	 */
	TINYINT, SMALLINT, INTEGER, BIGINT, FLOAT, REAL, DOUBLE, NUMERIC, DECIMAL,

	/**
	 * Time based
	 */
	DATE, TIME, TIMESTAMP,

	/**
	 * Booleans
	 */
	BIT, BOOLEAN,

	/**
	 * Other
	 */
	BINARY, VARBINARY, LONGVARBINARY, NULL, OTHER, JAVA_OBJECT, DISTINCT,
	STRUCT, ARRAY, BLOB, REF, DATALINK, ROWID, SQLXML;

	@SuppressWarnings("unchecked")
	public Comparator getComparator() {
		if (isTimeBased()) {
			return TimeComparator.getComparator();
		}
		if (isNumber()) {
			return NumberComparator.getComparator();
		}
		if (isLiteral()) {
			return ToStringComparator.getComparator();
		}
		return ObjectComparator.getComparator();
	}

	public boolean isBoolean() {
		switch (this) {
		case BIT:
		case BOOLEAN:
			return true;
		default:
			return false;
		}
	}

	public boolean isBinary() {
		switch (this) {
		case BIT:
		case BINARY:
		case VARBINARY:
		case LONGVARBINARY:
		case BLOB:
			return true;
		default:
			return false;
		}
	}

	public boolean isNumber() {
		switch (this) {
		case SMALLINT:
		case INTEGER:
		case BIGINT:
		case FLOAT:
		case REAL:
		case DOUBLE:
		case NUMERIC:
		case DECIMAL:
			return true;
		default:
			return false;
		}
	}

	public boolean isTimeBased() {
		switch (this) {
		case DATE:
		case TIME:
		case TIMESTAMP:
			return true;
		default:
			return false;
		}
	}

	public boolean isLiteral() {
		switch (this) {
		case CHAR:
		case NCHAR:
		case VARCHAR:
		case NVARCHAR:
		case LONGVARCHAR:
		case LONGNVARCHAR:
		case CLOB:
		case NCLOB:
			return true;
		default:
			return false;
		}
	}

	public boolean isLargeObject() {
		switch (this) {
		case BLOB:
		case CLOB:
		case NCLOB:
			return true;
		default:
			return false;
		}
	}

	/**
	 * Finds the ColumnType enum corresponding to the incoming JDBC
	 * type-constant
	 */
	public static ColumnType convertColumnType(int jdbcType) {
		try {
			Field[] fields = JdbcTypes.class.getFields();
			// We assume that the JdbcTypes class only consists of constant
			// integer types, so we make no assertions here
			for (int i = 0; i < fields.length; i++) {
				Field field = fields[i];
				int value = (Integer) field.getInt(null);
				if (value == jdbcType) {
					String fieldName = field.getName();
					ColumnType[] enumConstants = ColumnType.class
							.getEnumConstants();
					for (int j = 0; j < enumConstants.length; j++) {
						ColumnType columnType = enumConstants[j];
						if (fieldName.equals(columnType.toString())) {
							return columnType;
						}
					}
				}
			}
		} catch (Exception e) {
			throw new IllegalStateException(
					"Could not access fields in JdbcTypes", e);
		}
		return OTHER;
	}
}