/*
 * Copyright 2008 eobjects.dk
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.eobjects.metamodel.util;

import java.sql.Time;
import java.util.Comparator;

import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;

/**
 * General purpose comparator to use for objects of various kinds. Prevents
 * NullPointerExceptions and tries to use comparable interface if available and
 * appropriate on incoming objects.
 */
public class ObjectComparator implements Comparator<Object> {

	private static Comparator<Object> _instance = new ObjectComparator();

	public static Comparator<Object> getComparator() {
		return _instance;
	}

	private ObjectComparator() {
	}

	public static Comparable<Object> getComparable(final Object o) {
		return new Comparable<Object>() {

			public int compareTo(Object o2) {
				return _instance.compare(o, o2);
			}

			@Override
			public String toString() {
				return "ObjectComparable[object=" + o + "]";
			}
		};
	}

	@SuppressWarnings("unchecked")
	public int compare(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return -1;
		}
		if (o1 == null) {
			return -1;
		}
		if (o2 == null) {
			return 1;
		}
		if (o1 instanceof Comparable && o2 instanceof Comparable) {
			Comparable c1 = (Comparable) o1;
			Comparable c2 = (Comparable) o2;
			// We can only count on using the comparable interface if o1 and o2
			// are within of the same class or if one is a subclass of the other
			if (c1.getClass().isAssignableFrom(c2.getClass())) {
				return c1.compareTo(o2);
			}
			if (o2.getClass().isAssignableFrom(c1.getClass())) {
				return -1 * c2.compareTo(o1);
			}
		}
		if (o1 instanceof Number || o1 instanceof Number) {
			return NumberComparator.getComparator().compare(o1, o2);
		}
		if (isTimeBased(o1) || isTimeBased(o2)) {
			return TimeComparator.getComparator().compare(o1, o2);
		}
		return ToStringComparator.getComparator().compare(o1, o2);
	}

	private boolean isTimeBased(Object o) {
		if (o instanceof java.util.Date || o instanceof java.sql.Date
				|| o instanceof Time || o instanceof DateTime
				|| o instanceof LocalDate || o instanceof LocalTime) {
			return true;
		}
		return false;
	}

}
